@@ -3,45 +3,38 @@
 ** Copyright (C) 2012 Denis Shienkov <denis.shienkov@gmail.com>
 ** Copyright (C) 2012 Laszlo Papp <lpapp@kde.org>
 ** Copyright (C) 2012 Andre Hartmann <aha_1980@gmx.de>
-** Contact: http://www.qt-project.org/legal
+** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the QtSerialPort module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:LGPL21$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see http://www.qt.io/terms-conditions. For further
+** information use the contact form at http://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 2.1 or version 3 as published by the Free
+** Software Foundation and appearing in the file LICENSE.LGPLv21 and
+** LICENSE.LGPLv3 included in the packaging of this file. Please review the
+** following information to ensure the GNU Lesser General Public License
+** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
+** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
+** As a special exception, The Qt Company gives you certain additional
+** rights. These rights are described in The Qt Company LGPL Exception
 ** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
 **
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
-**
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
 
-#include "qserialport_unix_p.h"
+#include "qserialport_p.h"
+#include "qserialportinfo_p.h"
 
 #include <errno.h>
 #include <sys/time.h>
@@ -50,7 +43,7 @@
 #include <unistd.h>
 
 #ifdef Q_OS_MAC
-#if defined (MAC_OS_X_VERSION_10_4) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_4)
+#if defined(MAC_OS_X_VERSION_10_4) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_4)
 #include <IOKit/serial/ioss.h>
 #endif
 #endif
@@ -65,6 +58,10 @@
 #include <QtCore/qsocketnotifier.h>
 #include <QtCore/qmap.h>
 
+#ifdef Q_OS_MAC
+#include <QtCore/qstandardpaths.h>
+#endif
+
 QT_BEGIN_NAMESPACE
 
 QString serialPortLockFilePath(const QString &portName)
@@ -75,18 +72,33 @@ QString serialPortLockFilePath(const QString &portName)
         << QStringLiteral("/var/spool/locks")
         << QStringLiteral("/var/spool/uucp")
         << QStringLiteral("/tmp")
+        << QStringLiteral("/var/tmp")
+        << QStringLiteral("/var/lock/lockdev")
+        << QStringLiteral("/run/lock")
 #ifdef Q_OS_ANDROID
         << QStringLiteral("/data/local/tmp")
+#elif defined(Q_OS_OSX)
+           // This is the workaround to specify a temporary directory
+           // on OSX when running the App Sandbox feature.
+        << QStandardPaths::writableLocation(QStandardPaths::TempLocation);
 #endif
     ;
 
+    QString fileName = portName;
+    fileName.replace(QLatin1Char('/'), QLatin1Char('_'));
+    fileName.prepend(QLatin1String("/LCK.."));
+
     QString lockFilePath;
 
     foreach (const QString &lockDirectoryPath, lockDirectoryPaths) {
+        const QString filePath = lockDirectoryPath + fileName;
+
         QFileInfo lockDirectoryInfo(lockDirectoryPath);
-        if (lockDirectoryInfo.isReadable() && lockDirectoryInfo.isWritable()) {
-            lockFilePath = lockDirectoryPath;
-            break;
+        if (lockDirectoryInfo.isReadable()) {
+            if (QFile::exists(filePath) || lockDirectoryInfo.isWritable()) {
+                lockFilePath = filePath;
+                break;
+            }
         }
     }
 
@@ -97,29 +109,26 @@ QString serialPortLockFilePath(const QString &portName)
         return QString();
     }
 
-    QString replacedPortName = portName;
-
-    lockFilePath.append(QStringLiteral("/LCK.."));
-    lockFilePath.append(replacedPortName.replace(QLatin1Char('/'), QLatin1Char('_')));
-
     return lockFilePath;
 }
 
 class ReadNotifier : public QSocketNotifier
 {
-    Q_OBJECT
 public:
     ReadNotifier(QSerialPortPrivate *d, QObject *parent)
         : QSocketNotifier(d->descriptor, QSocketNotifier::Read, parent)
         , dptr(d)
-    {}
+    {
+    }
 
 protected:
-    bool event(QEvent *e) Q_DECL_OVERRIDE {
-        bool ret = QSocketNotifier::event(e);
-        if (ret)
+    bool event(QEvent *e) Q_DECL_OVERRIDE
+    {
+        if (e->type() == QEvent::SockAct) {
             dptr->readNotification();
-        return ret;
+            return true;
+        }
+        return QSocketNotifier::event(e);
     }
 
 private:
@@ -128,58 +137,41 @@ private:
 
 class WriteNotifier : public QSocketNotifier
 {
-    Q_OBJECT
 public:
     WriteNotifier(QSerialPortPrivate *d, QObject *parent)
         : QSocketNotifier(d->descriptor, QSocketNotifier::Write, parent)
         , dptr(d)
-    {}
+    {
+    }
 
 protected:
-    bool event(QEvent *e) Q_DECL_OVERRIDE {
-        bool ret = QSocketNotifier::event(e);
-        if (ret)
+    bool event(QEvent *e) Q_DECL_OVERRIDE
+    {
+        if (e->type() == QEvent::SockAct) {
             dptr->completeAsyncWrite();
-        return ret;
+            return true;
+        }
+        return QSocketNotifier::event(e);
     }
 
 private:
     QSerialPortPrivate *dptr;
 };
 
-#include "qserialport_unix.moc"
-
-QSerialPortPrivate::QSerialPortPrivate(QSerialPort *q)
-    : QSerialPortPrivateData(q)
-    , descriptor(-1)
-    , readNotifier(0)
-    , writeNotifier(0)
-    , readPortNotifierCalled(false)
-    , readPortNotifierState(false)
-    , readPortNotifierStateSet(false)
-    , emittedReadyRead(false)
-    , emittedBytesWritten(false)
-    , pendingBytesWritten(0)
-    , writeSequenceStarted(false)
-{
-}
-
 bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 {
-    Q_Q(QSerialPort);
-
-    QString lockFilePath = serialPortLockFilePath(portNameFromSystemLocation(systemLocation));
+    QString lockFilePath = serialPortLockFilePath(QSerialPortInfoPrivate::portNameFromSystemLocation(systemLocation));
     bool isLockFileEmpty = lockFilePath.isEmpty();
     if (isLockFileEmpty) {
         qWarning("Failed to create a lock file for opening the device");
-        q->setError(QSerialPort::PermissionError);
+        setError(QSerialPortErrorInfo(QSerialPort::PermissionError, QSerialPort::tr("Permission error while creating lock file")));
         return false;
     }
 
     QScopedPointer<QLockFile> newLockFileScopedPointer(new QLockFile(lockFilePath));
 
     if (!newLockFileScopedPointer->tryLock()) {
-        q->setError(QSerialPort::PermissionError);
+        setError(QSerialPortErrorInfo(QSerialPort::PermissionError, QSerialPort::tr("Permission error while locking the device")));
         return false;
     }
 
@@ -200,7 +192,7 @@ bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
     descriptor = qt_safe_open(systemLocation.toLocal8Bit().constData(), flags);
 
     if (descriptor == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -216,34 +208,30 @@ bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 
 void QSerialPortPrivate::close()
 {
-    Q_Q(QSerialPort);
-
     if (settingsRestoredOnClose) {
         if (::tcsetattr(descriptor, TCSANOW, &restoredTermios) == -1)
-            q->setError(decodeSystemError());
+            setError(getSystemError());
     }
 
 #ifdef TIOCNXCL
     if (::ioctl(descriptor, TIOCNXCL) == -1)
-        q->setError(decodeSystemError());
+        setError(getSystemError());
 #endif
 
     if (readNotifier) {
-        readNotifier->setEnabled(false);
-        readNotifier->deleteLater();
-        readNotifier = 0;
+        delete readNotifier;
+        readNotifier = Q_NULLPTR;
     }
 
     if (writeNotifier) {
-        writeNotifier->setEnabled(false);
-        writeNotifier->deleteLater();
-        writeNotifier = 0;
+        delete writeNotifier;
+        writeNotifier = Q_NULLPTR;
     }
 
     if (qt_safe_close(descriptor) == -1)
-        q->setError(decodeSystemError());
+        setError(getSystemError());
 
-    lockFileScopedPointer.reset(0);
+    lockFileScopedPointer.reset(Q_NULLPTR);
 
     descriptor = -1;
     pendingBytesWritten = 0;
@@ -252,12 +240,10 @@ void QSerialPortPrivate::close()
 
 QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
 {
-    Q_Q(QSerialPort);
-
     int arg = 0;
 
     if (::ioctl(descriptor, TIOCMGET, &arg) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return QSerialPort::NoSignal;
     }
 
@@ -290,14 +276,14 @@ QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
 #ifdef TIOCM_CAR
     if (arg & TIOCM_CAR)
         ret |= QSerialPort::DataCarrierDetectSignal;
-#elif defined TIOCM_CD
+#elif defined(TIOCM_CD)
     if (arg & TIOCM_CD)
         ret |= QSerialPort::DataCarrierDetectSignal;
 #endif
 #ifdef TIOCM_RNG
     if (arg & TIOCM_RNG)
         ret |= QSerialPort::RingIndicatorSignal;
-#elif defined TIOCM_RI
+#elif defined(TIOCM_RI)
     if (arg & TIOCM_RI)
         ret |= QSerialPort::RingIndicatorSignal;
 #endif
@@ -311,11 +297,9 @@ QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
 
 bool QSerialPortPrivate::setDataTerminalReady(bool set)
 {
-    Q_Q(QSerialPort);
-
     int status = TIOCM_DTR;
     if (::ioctl(descriptor, set ? TIOCMBIS : TIOCMBIC, &status) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -324,11 +308,9 @@ bool QSerialPortPrivate::setDataTerminalReady(bool set)
 
 bool QSerialPortPrivate::setRequestToSend(bool set)
 {
-    Q_Q(QSerialPort);
-
     int status = TIOCM_RTS;
     if (::ioctl(descriptor, set ? TIOCMBIS : TIOCMBIC, &status) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -342,11 +324,9 @@ bool QSerialPortPrivate::flush()
 
 bool QSerialPortPrivate::clear(QSerialPort::Directions directions)
 {
-    Q_Q(QSerialPort);
-
     if (::tcflush(descriptor, (directions == QSerialPort::AllDirections)
                      ? TCIOFLUSH : (directions & QSerialPort::Input) ? TCIFLUSH : TCOFLUSH) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -355,10 +335,8 @@ bool QSerialPortPrivate::clear(QSerialPort::Directions directions)
 
 bool QSerialPortPrivate::sendBreak(int duration)
 {
-    Q_Q(QSerialPort);
-
     if (::tcsendbreak(descriptor, duration) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -367,38 +345,24 @@ bool QSerialPortPrivate::sendBreak(int duration)
 
 bool QSerialPortPrivate::setBreakEnabled(bool set)
 {
-    Q_Q(QSerialPort);
-
     if (::ioctl(descriptor, set ? TIOCSBRK : TIOCCBRK) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
     return true;
 }
 
-void QSerialPortPrivate::startWriting()
-{
-    if (!isWriteNotificationEnabled())
-        setWriteNotificationEnabled(true);
-}
-
 bool QSerialPortPrivate::waitForReadyRead(int msecs)
 {
-    Q_Q(QSerialPort);
-
     QElapsedTimer stopWatch;
-
     stopWatch.start();
 
     do {
         bool readyToRead = false;
         bool readyToWrite = false;
-        bool timedOut = false;
         if (!waitForReadOrWrite(&readyToRead, &readyToWrite, true, !writeBuffer.isEmpty(),
-                                timeoutValue(msecs, stopWatch.elapsed()), &timedOut)) {
-            if (!timedOut)
-                q->setError(decodeSystemError());
+                                timeoutValue(msecs, stopWatch.elapsed()))) {
             return false;
         }
 
@@ -413,23 +377,17 @@ bool QSerialPortPrivate::waitForReadyRead(int msecs)
 
 bool QSerialPortPrivate::waitForBytesWritten(int msecs)
 {
-    Q_Q(QSerialPort);
-
     if (writeBuffer.isEmpty() && pendingBytesWritten <= 0)
         return false;
 
     QElapsedTimer stopWatch;
-
     stopWatch.start();
 
     forever {
         bool readyToRead = false;
         bool readyToWrite = false;
-        bool timedOut = false;
         if (!waitForReadOrWrite(&readyToRead, &readyToWrite, true, !writeBuffer.isEmpty(),
-                                timeoutValue(msecs, stopWatch.elapsed()), &timedOut)) {
-            if (!timedOut)
-                q->setError(decodeSystemError());
+                                timeoutValue(msecs, stopWatch.elapsed()))) {
             return false;
         }
 
@@ -451,32 +409,34 @@ bool QSerialPortPrivate::setBaudRate()
         && setBaudRate(outputBaudRate, QSerialPort::Output));
 }
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setBaudRate_helper(qint32 baudRate,
         QSerialPort::Directions directions)
 {
     if ((directions & QSerialPort::Input) && ::cfsetispeed(&currentTermios, baudRate) < 0)
-            return decodeSystemError();
+        return getSystemError();
 
     if ((directions & QSerialPort::Output) && ::cfsetospeed(&currentTermios, baudRate) < 0)
-            return decodeSystemError();
+        return getSystemError();
 
-    return QSerialPort::NoError;
+    return QSerialPortErrorInfo(QSerialPort::NoError);
 }
 
 #if defined(Q_OS_LINUX)
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setStandardBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     struct serial_struct currentSerialInfo;
 
+    ::memset(&currentSerialInfo, 0, sizeof(currentSerialInfo));
+
     if ((::ioctl(descriptor, TIOCGSERIAL, &currentSerialInfo) != -1)
             && (currentSerialInfo.flags & ASYNC_SPD_CUST)) {
         currentSerialInfo.flags &= ~ASYNC_SPD_CUST;
         currentSerialInfo.custom_divisor = 0;
         if (::ioctl(descriptor, TIOCSSERIAL, &currentSerialInfo) == -1)
-            return decodeSystemError();
+            return getSystemError();
     }
 
     return setBaudRate_helper(baudRate, directions);
@@ -484,7 +444,7 @@ QSerialPortPrivate::setStandardBaudRate(qint32 baudRate, QSerialPort::Directions
 
 #else
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setStandardBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     return setBaudRate_helper(baudRate, directions);
@@ -494,52 +454,58 @@ QSerialPortPrivate::setStandardBaudRate(qint32 baudRate, QSerialPort::Directions
 
 #if defined(Q_OS_LINUX)
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setCustomBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     Q_UNUSED(directions);
 
     struct serial_struct currentSerialInfo;
 
-    if (::ioctl(descriptor, TIOCGSERIAL, &currentSerialInfo) == -1)
-        return decodeSystemError();
+    ::memset(&currentSerialInfo, 0, sizeof(currentSerialInfo));
 
-    if (currentSerialInfo.baud_base % baudRate != 0)
-        return QSerialPort::UnsupportedOperationError;
+    if (::ioctl(descriptor, TIOCGSERIAL, &currentSerialInfo) == -1)
+        return getSystemError();
 
     currentSerialInfo.flags &= ~ASYNC_SPD_MASK;
     currentSerialInfo.flags |= (ASYNC_SPD_CUST /* | ASYNC_LOW_LATENCY*/);
     currentSerialInfo.custom_divisor = currentSerialInfo.baud_base / baudRate;
 
     if (currentSerialInfo.custom_divisor == 0)
-        return QSerialPort::UnsupportedOperationError;
+        return QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, QSerialPort::tr("No suitable custom baud rate divisor"));
+
+    if (currentSerialInfo.custom_divisor * baudRate != currentSerialInfo.baud_base) {
+        qWarning("Baud rate of serial port %s is set to %d instead of %d: divisor %f unsupported",
+            qPrintable(systemLocation),
+            currentSerialInfo.baud_base / currentSerialInfo.custom_divisor,
+            baudRate, (float)currentSerialInfo.baud_base / baudRate);
+    }
 
     if (::ioctl(descriptor, TIOCSSERIAL, &currentSerialInfo) == -1)
-        return decodeSystemError();
+        return getSystemError();
 
     return setBaudRate_helper(B38400, directions);
 }
 
 #elif defined(Q_OS_MAC)
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setCustomBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     Q_UNUSED(directions);
 
-#if defined (MAC_OS_X_VERSION_10_4) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_4)
+#if defined(MAC_OS_X_VERSION_10_4) && (MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_4)
     if (::ioctl(descriptor, IOSSIOSPEED, &baudRate) == -1)
-        return decodeSystemError();
+        return getSystemError();
 
-    return QSerialPort::NoError;
+    return QSerialPortErrorInfo(QSerialPort::NoError);
 #endif
 
-    return QSerialPort::UnsupportedOperationError;
+    return QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, QSerialPort::tr("Custom baud rate is not supported"));
 }
 
-#elif defined (Q_OS_QNX)
+#elif defined(Q_OS_QNX)
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setCustomBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     // On QNX, the values of the 'Bxxxx' constants are set to 'xxxx' (i.e.
@@ -551,35 +517,34 @@ QSerialPortPrivate::setCustomBaudRate(qint32 baudRate, QSerialPort::Directions d
 
 #else
 
-QSerialPort::SerialPortError
+QSerialPortErrorInfo
 QSerialPortPrivate::setCustomBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
     Q_UNUSED(baudRate);
     Q_UNUSED(directions);
 
-    return QSerialPort::UnsupportedOperationError;
+    return QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError);
 }
 
 #endif
 
 bool QSerialPortPrivate::setBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
-    Q_Q(QSerialPort);
-
     if (baudRate <= 0) {
-        q->setError(QSerialPort::UnsupportedOperationError);
+        setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, QSerialPort::tr("Invalid baud rate value")));
         return false;
     }
 
     const qint32 unixBaudRate = QSerialPortPrivate::settingFromBaudRate(baudRate);
 
-    const QSerialPort::SerialPortError error = (unixBaudRate > 0)
+    const QSerialPortErrorInfo error = (unixBaudRate > 0)
         ? setStandardBaudRate(unixBaudRate, directions)
         : setCustomBaudRate(baudRate, directions);
 
-    if (error == QSerialPort::NoError)
+    if (error.errorCode == QSerialPort::NoError)
         return updateTermios();
 
+    setError(error);
     return false;
 }
 
@@ -719,23 +684,12 @@ bool QSerialPortPrivate::readNotification()
 {
     Q_Q(QSerialPort);
 
-    // Prevent recursive calls
-    if (readPortNotifierCalled) {
-        if (!readPortNotifierStateSet) {
-            readPortNotifierStateSet = true;
-            readPortNotifierState = isReadNotificationEnabled();
-            setReadNotificationEnabled(false);
-        }
-    }
-
-    readPortNotifierCalled = true;
-
     // Always buffered, read data from the port into the read buffer
-    qint64 newBytes = readBuffer.size();
+    qint64 newBytes = buffer.size();
     qint64 bytesToRead = policy == QSerialPort::IgnorePolicy ? ReadChunkSize : 1;
 
-    if (readBufferMaxSize && bytesToRead > (readBufferMaxSize - readBuffer.size())) {
-        bytesToRead = readBufferMaxSize - readBuffer.size();
+    if (readBufferMaxSize && bytesToRead > (readBufferMaxSize - buffer.size())) {
+        bytesToRead = readBufferMaxSize - buffer.size();
         if (bytesToRead == 0) {
             // Buffer is full. User must read data from the buffer
             // before we can read more from the port.
@@ -743,26 +697,26 @@ bool QSerialPortPrivate::readNotification()
         }
     }
 
-    char *ptr = readBuffer.reserve(bytesToRead);
+    char *ptr = buffer.reserve(bytesToRead);
     const qint64 readBytes = readFromPort(ptr, bytesToRead);
 
     if (readBytes <= 0) {
-        QSerialPort::SerialPortError error = decodeSystemError();
-        if (error != QSerialPort::ResourceError)
-            error = QSerialPort::ReadError;
+        QSerialPortErrorInfo error = getSystemError();
+        if (error.errorCode != QSerialPort::ResourceError)
+            error.errorCode = QSerialPort::ReadError;
         else
             setReadNotificationEnabled(false);
-        q->setError(error);
-        readBuffer.chop(bytesToRead);
+        setError(error);
+        buffer.chop(bytesToRead);
         return false;
     }
 
-    readBuffer.chop(bytesToRead - qMax(readBytes, qint64(0)));
+    buffer.chop(bytesToRead - qMax(readBytes, qint64(0)));
 
-    newBytes = readBuffer.size() - newBytes;
+    newBytes = buffer.size() - newBytes;
 
     // If read buffer is full, disable the read port notifier.
-    if (readBufferMaxSize && readBuffer.size() == readBufferMaxSize)
+    if (readBufferMaxSize && buffer.size() == readBufferMaxSize)
         setReadNotificationEnabled(false);
 
     // only emit readyRead() when not recursing, and only if there is data available
@@ -774,33 +728,21 @@ bool QSerialPortPrivate::readNotification()
         emittedReadyRead = false;
     }
 
-    if (!hasData)
-        setReadNotificationEnabled(true);
-
-    // reset the read port notifier state if we reentered inside the
-    // readyRead() connected slot.
-    if (readPortNotifierStateSet
-            && readPortNotifierState != isReadNotificationEnabled()) {
-        setReadNotificationEnabled(readPortNotifierState);
-        readPortNotifierStateSet = false;
-    }
     return true;
 }
 
 bool QSerialPortPrivate::startAsyncWrite()
 {
-    Q_Q(QSerialPort);
-
     if (writeBuffer.isEmpty() || writeSequenceStarted)
         return true;
 
     // Attempt to write it all in one chunk.
     qint64 written = writeToPort(writeBuffer.readPointer(), writeBuffer.nextDataBlockSize());
     if (written < 0) {
-        QSerialPort::SerialPortError error = decodeSystemError();
-        if (error != QSerialPort::ResourceError)
-            error = QSerialPort::WriteError;
-        q->setError(error);
+        QSerialPortErrorInfo error = getSystemError();
+        if (error.errorCode != QSerialPort::ResourceError)
+            error.errorCode = QSerialPort::WriteError;
+        setError(error);
         return false;
     }
 
@@ -838,15 +780,13 @@ bool QSerialPortPrivate::completeAsyncWrite()
 
 inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
 {
-    Q_Q(QSerialPort);
-
 #ifdef TIOCEXCL
     if (::ioctl(descriptor, TIOCEXCL) == -1)
-        q->setError(decodeSystemError());
+        setError(getSystemError());
 #endif
 
     if (::tcgetattr(descriptor, &restoredTermios) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -876,51 +816,82 @@ inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
     return true;
 }
 
+qint64 QSerialPortPrivate::writeData(const char *data, qint64 maxSize)
+{
+    ::memcpy(writeBuffer.reserve(maxSize), data, maxSize);
+    if (!writeBuffer.isEmpty() && !isWriteNotificationEnabled())
+        setWriteNotificationEnabled(true);
+    return maxSize;
+}
+
 bool QSerialPortPrivate::updateTermios()
 {
-    Q_Q(QSerialPort);
-
     if (::tcsetattr(descriptor, TCSANOW, &currentTermios) == -1) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
     return true;
 }
 
-QSerialPort::SerialPortError QSerialPortPrivate::decodeSystemError() const
+QSerialPortErrorInfo QSerialPortPrivate::getSystemError(int systemErrorCode) const
 {
-    QSerialPort::SerialPortError error;
-    switch (errno) {
+    if (systemErrorCode == -1)
+        systemErrorCode = errno;
+
+    QSerialPortErrorInfo error;
+    error.errorString = qt_error_string(systemErrorCode);
+
+    switch (systemErrorCode) {
     case ENODEV:
-        error = QSerialPort::DeviceNotFoundError;
+        error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
 #ifdef ENOENT
     case ENOENT:
-        error = QSerialPort::DeviceNotFoundError;
+        error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
 #endif
     case EACCES:
-        error = QSerialPort::PermissionError;
+        error.errorCode = QSerialPort::PermissionError;
         break;
     case EBUSY:
-        error = QSerialPort::PermissionError;
+        error.errorCode = QSerialPort::PermissionError;
         break;
     case EAGAIN:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
     case EIO:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
     case EBADF:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
 #ifdef Q_OS_MAC
     case ENXIO:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
+        break;
+#endif
+#ifdef EINVAL
+    case EINVAL:
+        error.errorCode = QSerialPort::UnsupportedOperationError;
+        break;
+#endif
+#ifdef ENOIOCTLCMD
+    case ENOIOCTLCMD:
+        error.errorCode = QSerialPort::UnsupportedOperationError;
+        break;
+#endif
+#ifdef ENOTTY
+    case ENOTTY:
+        error.errorCode = QSerialPort::UnsupportedOperationError;
+        break;
+#endif
+#ifdef EPERM
+    case EPERM:
+        error.errorCode = QSerialPort::PermissionError;
         break;
 #endif
     default:
-        error = QSerialPort::UnknownError;
+        error.errorCode = QSerialPort::UnknownError;
         break;
     }
     return error;
@@ -962,13 +933,10 @@ void QSerialPortPrivate::setWriteNotificationEnabled(bool enable)
 
 bool QSerialPortPrivate::waitForReadOrWrite(bool *selectForRead, bool *selectForWrite,
                                            bool checkRead, bool checkWrite,
-                                           int msecs, bool *timedOut)
+                                           int msecs)
 {
-    Q_Q(QSerialPort);
-
     Q_ASSERT(selectForRead);
     Q_ASSERT(selectForWrite);
-    Q_ASSERT(timedOut);
 
     fd_set fdread;
     FD_ZERO(&fdread);
@@ -984,25 +952,25 @@ bool QSerialPortPrivate::waitForReadOrWrite(bool *selectForRead, bool *selectFor
     tv.tv_sec = msecs / 1000;
     tv.tv_usec = (msecs % 1000) * 1000;
 
-    int ret = ::select(descriptor + 1, &fdread, &fdwrite, 0, msecs < 0 ? 0 : &tv);
-    if (ret < 0)
+    const int ret = ::select(descriptor + 1, &fdread, &fdwrite, 0, msecs < 0 ? 0 : &tv);
+    if (ret < 0) {
+        setError(getSystemError());
         return false;
+    }
     if (ret == 0) {
-        *timedOut = true;
-        q->setError(QSerialPort::TimeoutError);
+        setError(QSerialPortErrorInfo(QSerialPort::TimeoutError, QSerialPort::tr("Operation timed out")));
         return false;
     }
 
     *selectForRead = FD_ISSET(descriptor, &fdread);
     *selectForWrite = FD_ISSET(descriptor, &fdwrite);
-
-    return ret;
+    return true;
 }
 
 qint64 QSerialPortPrivate::readFromPort(char *data, qint64 maxSize)
 {
     qint64 bytesRead = 0;
-#if defined (CMSPAR)
+#if defined(CMSPAR)
     if (parity == QSerialPort::NoParity
             || policy != QSerialPort::StopReceivingPolicy) {
 #else
@@ -1020,7 +988,7 @@ qint64 QSerialPortPrivate::readFromPort(char *data, qint64 maxSize)
 qint64 QSerialPortPrivate::writeToPort(const char *data, qint64 maxSize)
 {
     qint64 bytesWritten = 0;
-#if defined (CMSPAR)
+#if defined(CMSPAR)
     bytesWritten = qt_safe_write(descriptor, data, maxSize);
 #else
     if (parity != QSerialPort::MarkParity
@@ -1077,8 +1045,6 @@ qint64 QSerialPortPrivate::writePerChar(const char *data, qint64 maxSize)
 
 qint64 QSerialPortPrivate::readPerChar(char *data, qint64 maxSize)
 {
-    Q_Q(QSerialPort);
-
     qint64 ret = 0;
     quint8 const charMask = (0xFF >> (8 - dataBits));
 
@@ -1124,13 +1090,15 @@ qint64 QSerialPortPrivate::readPerChar(char *data, qint64 maxSize)
             switch (policy) {
             case QSerialPort::SkipPolicy:
                 continue;       //ignore received character
-            case QSerialPort::StopReceivingPolicy:
+            case QSerialPort::StopReceivingPolicy: {
                 if (parity != QSerialPort::NoParity)
-                    q->setError(QSerialPort::ParityError);
+                    setError(QSerialPortErrorInfo(QSerialPort::ParityError, QSerialPort::tr("Parity error detected while reading")));
+                else if (*data == '\0')
+                    setError(QSerialPortErrorInfo(QSerialPort::BreakConditionError, QSerialPort::tr("Break condition detected while reading")));
                 else
-                    q->setError(*data == '\0' ?
-                                QSerialPort::BreakConditionError : QSerialPort::FramingError);
+                    setError(QSerialPortErrorInfo(QSerialPort::FramingError, QSerialPort::tr("Framing error detected while reading")));
                 return ++ret;   //abort receiving
+            }
                 break;
             case QSerialPort::UnknownPolicy:
                 // Unknown error policy is used! Falling back to PassZeroPolicy
@@ -1147,38 +1115,6 @@ qint64 QSerialPortPrivate::readPerChar(char *data, qint64 maxSize)
     return ret;
 }
 
-#ifdef Q_OS_MAC
-static const QString defaultFilePathPrefix = QStringLiteral("/dev/cu.");
-static const QString unusedFilePathPrefix = QStringLiteral("/dev/tty.");
-#else
-static const QString defaultFilePathPrefix = QStringLiteral("/dev/");
-#endif
-
-QString QSerialPortPrivate::portNameToSystemLocation(const QString &port)
-{
-    QString ret = port;
-
-#ifdef Q_OS_MAC
-    ret.remove(unusedFilePathPrefix);
-#endif
-
-    if (!ret.contains(defaultFilePathPrefix))
-        ret.prepend(defaultFilePathPrefix);
-    return ret;
-}
-
-QString QSerialPortPrivate::portNameFromSystemLocation(const QString &location)
-{
-    QString ret = location;
-
-#ifdef Q_OS_MAC
-    ret.remove(unusedFilePathPrefix);
-#endif
-
-    ret.remove(defaultFilePathPrefix);
-    return ret;
-}
-
 typedef QMap<qint32, qint32> BaudRateMap;
 
 // The OS specific defines can be found in termios.h
