@@ -3,51 +3,43 @@
 ** Copyright (C) 2012 Denis Shienkov <denis.shienkov@gmail.com>
 ** Copyright (C) 2012 Laszlo Papp <lpapp@kde.org>
 ** Copyright (C) 2012 Andre Hartmann <aha_1980@gmx.de>
-** Contact: http://www.qt-project.org/legal
+** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the QtSerialPort module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:LGPL21$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see http://www.qt.io/terms-conditions. For further
+** information use the contact form at http://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 2.1 or version 3 as published by the Free
+** Software Foundation and appearing in the file LICENSE.LGPLv21 and
+** LICENSE.LGPLv3 included in the packaging of this file. Please review the
+** following information to ensure the GNU Lesser General Public License
+** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
+** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
+** As a special exception, The Qt Company gives you certain additional
+** rights. These rights are described in The Qt Company LGPL Exception
 ** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
 **
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
-**
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
 
-#include "qserialport_win_p.h"
+#include "qserialport_p.h"
 
 #include <QtCore/qcoreevent.h>
 #include <QtCore/qelapsedtimer.h>
 #include <QtCore/qvector.h>
 #include <QtCore/qtimer.h>
-#include <QtCore/qwineventnotifier.h>
+#include <private/qwinoverlappedionotifier_p.h>
 #include <algorithm>
 
 #ifndef CTL_CODE
@@ -83,60 +75,8 @@
 
 QT_BEGIN_NAMESPACE
 
-static void initializeOverlappedStructure(OVERLAPPED &overlapped)
-{
-    overlapped.Internal = 0;
-    overlapped.InternalHigh = 0;
-    overlapped.Offset = 0;
-    overlapped.OffsetHigh = 0;
-}
-
-QSerialPortPrivate::QSerialPortPrivate(QSerialPort *q)
-    : QSerialPortPrivateData(q)
-    , handle(INVALID_HANDLE_VALUE)
-    , parityErrorOccurred(false)
-    , readChunkBuffer(ReadChunkSize, 0)
-    , readyReadEmitted(0)
-    , writeStarted(false)
-    , communicationNotifier(new QWinEventNotifier(q))
-    , readCompletionNotifier(new QWinEventNotifier(q))
-    , writeCompletionNotifier(new QWinEventNotifier(q))
-    , startAsyncWriteTimer(0)
-    , originalEventMask(0)
-    , triggeredEventMask(0)
-{
-    ::ZeroMemory(&communicationOverlapped, sizeof(communicationOverlapped));
-    communicationOverlapped.hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
-    if (!communicationOverlapped.hEvent)
-        q->setError(decodeSystemError());
-    else {
-        communicationNotifier->setHandle(communicationOverlapped.hEvent);
-        q->connect(communicationNotifier, SIGNAL(activated(HANDLE)), q, SLOT(_q_completeAsyncCommunication()));
-    }
-
-    ::ZeroMemory(&readCompletionOverlapped, sizeof(readCompletionOverlapped));
-    readCompletionOverlapped.hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
-    if (!readCompletionOverlapped.hEvent)
-        q->setError(decodeSystemError());
-    else {
-        readCompletionNotifier->setHandle(readCompletionOverlapped.hEvent);
-        q->connect(readCompletionNotifier, SIGNAL(activated(HANDLE)), q, SLOT(_q_completeAsyncRead()));
-    }
-
-    ::ZeroMemory(&writeCompletionOverlapped, sizeof(writeCompletionOverlapped));
-    writeCompletionOverlapped.hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
-    if (!writeCompletionOverlapped.hEvent)
-        q->setError(decodeSystemError());
-    else {
-        writeCompletionNotifier->setHandle(writeCompletionOverlapped.hEvent);
-        q->connect(writeCompletionNotifier, SIGNAL(activated(HANDLE)), q, SLOT(_q_completeAsyncWrite()));
-    }
-}
-
 bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 {
-    Q_Q(QSerialPort);
-
     DWORD desiredAccess = 0;
     originalEventMask = EV_ERR;
 
@@ -148,14 +88,14 @@ bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
         desiredAccess |= GENERIC_WRITE;
 
     handle = ::CreateFile(reinterpret_cast<const wchar_t*>(systemLocation.utf16()),
-                              desiredAccess, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
+                              desiredAccess, 0, Q_NULLPTR, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, Q_NULLPTR);
 
     if (handle == INVALID_HANDLE_VALUE) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
-    if (initialize(mode))
+    if (initialize())
         return true;
 
     ::CloseHandle(handle);
@@ -164,44 +104,45 @@ bool QSerialPortPrivate::open(QIODevice::OpenMode mode)
 
 void QSerialPortPrivate::close()
 {
-    Q_Q(QSerialPort);
-
     if (!::CancelIo(handle))
-        q->setError(decodeSystemError());
+        setError(getSystemError());
 
-    readCompletionNotifier->setEnabled(false);
-    writeCompletionNotifier->setEnabled(false);
-    communicationNotifier->setEnabled(false);
+    if (notifier) {
+        delete notifier;
+        notifier = Q_NULLPTR;
+    }
 
-    readBuffer.clear();
+    if (startAsyncWriteTimer) {
+        delete startAsyncWriteTimer;
+        startAsyncWriteTimer = Q_NULLPTR;
+    }
 
+    readStarted = false;
     writeStarted = false;
     writeBuffer.clear();
+    actualBytesToWrite = 0;
 
-    readyReadEmitted = false;
     parityErrorOccurred = false;
 
     if (settingsRestoredOnClose) {
         if (!::SetCommState(handle, &restoredDcb))
-            q->setError(decodeSystemError());
+            setError(getSystemError());
         else if (!::SetCommTimeouts(handle, &restoredCommTimeouts))
-            q->setError(decodeSystemError());
+            setError(getSystemError());
     }
 
     if (!::CloseHandle(handle))
-        q->setError(decodeSystemError());
+        setError(getSystemError());
 
     handle = INVALID_HANDLE_VALUE;
 }
 
 QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
 {
-    Q_Q(QSerialPort);
-
     DWORD modemStat = 0;
 
     if (!::GetCommModemStatus(handle, &modemStat)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return QSerialPort::NoSignal;
     }
 
@@ -217,10 +158,10 @@ QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
         ret |= QSerialPort::DataCarrierDetectSignal;
 
     DWORD bytesReturned = 0;
-    if (!::DeviceIoControl(handle, IOCTL_SERIAL_GET_DTRRTS, NULL, 0,
+    if (!::DeviceIoControl(handle, IOCTL_SERIAL_GET_DTRRTS, Q_NULLPTR, 0,
                           &modemStat, sizeof(modemStat),
-                          &bytesReturned, NULL)) {
-        q->setError(decodeSystemError());
+                          &bytesReturned, Q_NULLPTR)) {
+        setError(getSystemError());
         return ret;
     }
 
@@ -234,10 +175,8 @@ QSerialPort::PinoutSignals QSerialPortPrivate::pinoutSignals()
 
 bool QSerialPortPrivate::setDataTerminalReady(bool set)
 {
-    Q_Q(QSerialPort);
-
     if (!::EscapeCommFunction(handle, set ? SETDTR : CLRDTR)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -247,10 +186,8 @@ bool QSerialPortPrivate::setDataTerminalReady(bool set)
 
 bool QSerialPortPrivate::setRequestToSend(bool set)
 {
-    Q_Q(QSerialPort);
-
     if (!::EscapeCommFunction(handle, set ? SETRTS : CLRRTS)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -264,20 +201,24 @@ bool QSerialPortPrivate::flush()
 
 bool QSerialPortPrivate::clear(QSerialPort::Directions directions)
 {
-    Q_Q(QSerialPort);
-
     DWORD flags = 0;
     if (directions & QSerialPort::Input)
         flags |= PURGE_RXABORT | PURGE_RXCLEAR;
     if (directions & QSerialPort::Output) {
         flags |= PURGE_TXABORT | PURGE_TXCLEAR;
-        writeStarted = false;
+        actualBytesToWrite = 0;
     }
     if (!::PurgeComm(handle, flags)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
+    // We need start async read because a reading can be stalled. Since the
+    // PurgeComm can abort of current reading sequence, or a port is in hardware
+    // flow control mode, or a port has a limited read buffer size.
+    if (directions & QSerialPort::Input)
+        startAsyncCommunication();
+
     return true;
 }
 
@@ -296,74 +237,40 @@ bool QSerialPortPrivate::sendBreak(int duration)
 
 bool QSerialPortPrivate::setBreakEnabled(bool set)
 {
-    Q_Q(QSerialPort);
-
     if (set ? !::SetCommBreak(handle) : !::ClearCommBreak(handle)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
     return true;
 }
 
-void QSerialPortPrivate::startWriting()
-{
-    Q_Q(QSerialPort);
-
-    if (!writeStarted) {
-        if (!startAsyncWriteTimer) {
-            startAsyncWriteTimer = new QTimer(q);
-            q->connect(startAsyncWriteTimer, SIGNAL(timeout()), q, SLOT(_q_startAsyncWrite()));
-            startAsyncWriteTimer->setSingleShot(true);
-        }
-        startAsyncWriteTimer->start(0);
-    }
-}
-
 bool QSerialPortPrivate::waitForReadyRead(int msecs)
 {
-    Q_Q(QSerialPort);
+    if (!writeStarted && !_q_startAsyncWrite())
+        return false;
+
+    const qint64 initialReadBufferSize = buffer.size();
+    qint64 currentReadBufferSize = initialReadBufferSize;
 
     QElapsedTimer stopWatch;
     stopWatch.start();
 
-    if (!writeStarted && !_q_startAsyncWrite())
-        return false;
-
-    const qint64 initialReadBufferSize = readBuffer.size();
-    qint64 currentReadBufferSize = initialReadBufferSize;
-
     do {
-        bool timedOut = false;
-        HANDLE triggeredEvent = 0;
-
-        if (!waitAnyEvent(timeoutValue(msecs, stopWatch.elapsed()), &timedOut, &triggeredEvent) || !triggeredEvent) {
-            // This is occur timeout or another error
-            if (!timedOut)
-                q->setError(decodeSystemError());
+        OVERLAPPED *overlapped = waitForNotified(timeoutValue(msecs, stopWatch.elapsed()));
+        if (!overlapped)
             return false;
-        }
 
-        if (triggeredEvent == communicationOverlapped.hEvent) {
-            if (!_q_completeAsyncCommunication())
-                return false;
-        } else if (triggeredEvent == readCompletionOverlapped.hEvent) {
-            if (!_q_completeAsyncRead())
-                return false;
-            const qint64 readBytesForOneReadOperation = qint64(readBuffer.size()) - currentReadBufferSize;
+        if (overlapped == &readCompletionOverlapped) {
+            const qint64 readBytesForOneReadOperation = qint64(buffer.size()) - currentReadBufferSize;
             if (readBytesForOneReadOperation == ReadChunkSize) {
-                currentReadBufferSize = readBuffer.size();
+                currentReadBufferSize = buffer.size();
             } else if (readBytesForOneReadOperation == 0) {
                 if (initialReadBufferSize != currentReadBufferSize)
                     return true;
             } else {
                 return true;
             }
-        } else if (triggeredEvent == writeCompletionOverlapped.hEvent) {
-            if (!_q_completeAsyncWrite())
-                return false;
-        } else {
-            return false;
         }
 
     } while (msecs == -1 || timeoutValue(msecs, stopWatch.elapsed()) > 0);
@@ -373,39 +280,22 @@ bool QSerialPortPrivate::waitForReadyRead(int msecs)
 
 bool QSerialPortPrivate::waitForBytesWritten(int msecs)
 {
-    Q_Q(QSerialPort);
-
     if (writeBuffer.isEmpty())
         return false;
 
+    if (!writeStarted && !_q_startAsyncWrite())
+        return false;
+
     QElapsedTimer stopWatch;
     stopWatch.start();
 
-    if (!writeStarted && !_q_startAsyncWrite())
-        return false;
-
     forever {
-        bool timedOut = false;
-        HANDLE triggeredEvent = 0;
-
-        if (!waitAnyEvent(timeoutValue(msecs, stopWatch.elapsed()), &timedOut, &triggeredEvent) || !triggeredEvent) {
-            if (!timedOut)
-                q->setError(decodeSystemError());
+        OVERLAPPED *overlapped = waitForNotified(timeoutValue(msecs, stopWatch.elapsed()));
+        if (!overlapped)
             return false;
-        }
-
-        if (triggeredEvent == communicationOverlapped.hEvent) {
-            if (!_q_completeAsyncCommunication())
-                return false;
-        } else if (triggeredEvent == readCompletionOverlapped.hEvent) {
-            if (!_q_completeAsyncRead())
-                return false;
-        } else if (triggeredEvent == writeCompletionOverlapped.hEvent) {
-            return _q_completeAsyncWrite();
-        } else {
-            return false;
-        }
 
+         if (overlapped == &writeCompletionOverlapped)
+            return true;
     }
 
     return false;
@@ -418,10 +308,8 @@ bool QSerialPortPrivate::setBaudRate()
 
 bool QSerialPortPrivate::setBaudRate(qint32 baudRate, QSerialPort::Directions directions)
 {
-    Q_Q(QSerialPort);
-
     if (directions != QSerialPort::AllDirections) {
-        q->setError(QSerialPort::UnsupportedOperationError);
+        setError(QSerialPortErrorInfo(QSerialPort::UnsupportedOperationError, QSerialPort::tr("Custom baud rate direction is unsupported")));
         return false;
     }
     currentDcb.BaudRate = baudRate;
@@ -510,9 +398,9 @@ bool QSerialPortPrivate::setDataErrorPolicy(QSerialPort::DataErrorPolicy policy)
     return true;
 }
 
-bool QSerialPortPrivate::_q_completeAsyncCommunication()
+bool QSerialPortPrivate::completeAsyncCommunication(qint64 bytesTransferred)
 {
-    if (handleOverlappedResult(0, communicationOverlapped) == qint64(-1))
+    if (bytesTransferred == qint64(-1))
         return false;
     if (EV_ERR & triggeredEventMask)
         handleLineStatusErrors();
@@ -520,30 +408,36 @@ bool QSerialPortPrivate::_q_completeAsyncCommunication()
     return startAsyncRead();
 }
 
-bool QSerialPortPrivate::_q_completeAsyncRead()
+bool QSerialPortPrivate::completeAsyncRead(qint64 bytesTransferred)
 {
-    const qint64 bytesTransferred = handleOverlappedResult(QSerialPort::Input, readCompletionOverlapped);
-    if (bytesTransferred == qint64(-1))
+    if (bytesTransferred == qint64(-1)) {
+        readStarted = false;
         return false;
+    }
     if (bytesTransferred > 0) {
-        readBuffer.append(readChunkBuffer.left(bytesTransferred));
-        if (!emulateErrorPolicy())
-            emitReadyRead();
+        char *ptr = buffer.reserve(bytesTransferred);
+        ::memcpy(ptr, readChunkBuffer.constData(), bytesTransferred);
     }
 
-    // start async read for possible remainder into driver queue
+    readStarted = false;
+
+    bool result = true;
     if ((bytesTransferred == ReadChunkSize) && (policy == QSerialPort::IgnorePolicy))
-        return startAsyncRead();
-    else // driver queue is emplty, so startup wait comm event
-        return startAsyncCommunication();
+        result = startAsyncRead();
+    else if (readBufferMaxSize == 0 || readBufferMaxSize > buffer.size())
+        result = startAsyncCommunication();
+
+    if ((bytesTransferred > 0) && !emulateErrorPolicy())
+        emitReadyRead();
+
+    return result;
 }
 
-bool QSerialPortPrivate::_q_completeAsyncWrite()
+bool QSerialPortPrivate::completeAsyncWrite(qint64 bytesTransferred)
 {
     Q_Q(QSerialPort);
 
     if (writeStarted) {
-        const qint64 bytesTransferred = handleOverlappedResult(QSerialPort::Output, writeCompletionOverlapped);
         if (bytesTransferred == qint64(-1)) {
             writeStarted = false;
             return false;
@@ -559,15 +453,13 @@ bool QSerialPortPrivate::_q_completeAsyncWrite()
 
 bool QSerialPortPrivate::startAsyncCommunication()
 {
-    Q_Q(QSerialPort);
-
-    initializeOverlappedStructure(communicationOverlapped);
+    ::ZeroMemory(&communicationOverlapped, sizeof(communicationOverlapped));
     if (!::WaitCommEvent(handle, &triggeredEventMask, &communicationOverlapped)) {
-        QSerialPort::SerialPortError error = decodeSystemError();
-        if (error != QSerialPort::NoError) {
-            if (error == QSerialPort::PermissionError)
-                error = QSerialPort::ResourceError;
-            q->setError(error);
+        QSerialPortErrorInfo error = getSystemError();
+        if (error.errorCode != QSerialPort::NoError) {
+            if (error.errorCode == QSerialPort::PermissionError)
+                error.errorCode = QSerialPort::ResourceError;
+            setError(error);
             return false;
         }
     }
@@ -576,12 +468,13 @@ bool QSerialPortPrivate::startAsyncCommunication()
 
 bool QSerialPortPrivate::startAsyncRead()
 {
-    Q_Q(QSerialPort);
+    if (readStarted)
+        return true;
 
     DWORD bytesToRead = policy == QSerialPort::IgnorePolicy ? ReadChunkSize : 1;
 
-    if (readBufferMaxSize && bytesToRead > (readBufferMaxSize - readBuffer.size())) {
-        bytesToRead = readBufferMaxSize - readBuffer.size();
+    if (readBufferMaxSize && bytesToRead > (readBufferMaxSize - buffer.size())) {
+        bytesToRead = readBufferMaxSize - buffer.size();
         if (bytesToRead == 0) {
             // Buffer is full. User must read data from the buffer
             // before we can read more from the port.
@@ -589,48 +482,68 @@ bool QSerialPortPrivate::startAsyncRead()
         }
     }
 
-    initializeOverlappedStructure(readCompletionOverlapped);
-    if (::ReadFile(handle, readChunkBuffer.data(), bytesToRead, NULL, &readCompletionOverlapped))
+    ::ZeroMemory(&readCompletionOverlapped, sizeof(readCompletionOverlapped));
+    if (::ReadFile(handle, readChunkBuffer.data(), bytesToRead, Q_NULLPTR, &readCompletionOverlapped)) {
+        readStarted = true;
         return true;
+    }
 
-    QSerialPort::SerialPortError error = decodeSystemError();
-    if (error != QSerialPort::NoError) {
-        if (error == QSerialPort::PermissionError)
-            error = QSerialPort::ResourceError;
-        if (error != QSerialPort::ResourceError)
-            error = QSerialPort::ReadError;
-        q->setError(error);
+    QSerialPortErrorInfo error = getSystemError();
+    if (error.errorCode != QSerialPort::NoError) {
+        if (error.errorCode == QSerialPort::PermissionError)
+            error.errorCode = QSerialPort::ResourceError;
+        if (error.errorCode != QSerialPort::ResourceError)
+            error.errorCode = QSerialPort::ReadError;
+        setError(error);
         return false;
     }
 
+    readStarted = true;
     return true;
 }
 
 bool QSerialPortPrivate::_q_startAsyncWrite()
 {
-    Q_Q(QSerialPort);
-
     if (writeBuffer.isEmpty() || writeStarted)
         return true;
 
-    initializeOverlappedStructure(writeCompletionOverlapped);
+    const int writeBytes = writeBuffer.nextDataBlockSize();
+    ::ZeroMemory(&writeCompletionOverlapped, sizeof(writeCompletionOverlapped));
     if (!::WriteFile(handle, writeBuffer.readPointer(),
-                     writeBuffer.nextDataBlockSize(),
-                     NULL, &writeCompletionOverlapped)) {
+                     writeBytes, Q_NULLPTR, &writeCompletionOverlapped)) {
 
-        QSerialPort::SerialPortError error = decodeSystemError();
-        if (error != QSerialPort::NoError) {
-            if (error != QSerialPort::ResourceError)
-                error = QSerialPort::WriteError;
-            q->setError(error);
+        QSerialPortErrorInfo error = getSystemError();
+        if (error.errorCode != QSerialPort::NoError) {
+            if (error.errorCode != QSerialPort::ResourceError)
+                error.errorCode = QSerialPort::WriteError;
+            setError(error);
             return false;
         }
     }
 
+    actualBytesToWrite -= writeBytes;
     writeStarted = true;
     return true;
 }
 
+void QSerialPortPrivate::_q_notified(DWORD numberOfBytes, DWORD errorCode, OVERLAPPED *overlapped)
+{
+    const QSerialPortErrorInfo error = getSystemError(errorCode);
+    if (error.errorCode != QSerialPort::NoError) {
+        setError(error);
+        return;
+    }
+
+    if (overlapped == &communicationOverlapped)
+        completeAsyncCommunication(numberOfBytes);
+    else if (overlapped == &readCompletionOverlapped)
+        completeAsyncRead(numberOfBytes);
+    else if (overlapped == &writeCompletionOverlapped)
+        completeAsyncWrite(numberOfBytes);
+    else
+        Q_ASSERT(!"Unknown OVERLAPPED activated");
+}
+
 bool QSerialPortPrivate::emulateErrorPolicy()
 {
     if (!parityErrorOccurred)
@@ -640,11 +553,11 @@ bool QSerialPortPrivate::emulateErrorPolicy()
 
     switch (policy) {
     case QSerialPort::SkipPolicy:
-        readBuffer.getChar();
+        buffer.getChar();
         break;
     case QSerialPort::PassZeroPolicy:
-        readBuffer.getChar();
-        readBuffer.putChar('\0');
+        buffer.getChar();
+        buffer.ungetChar('\0');
         emitReadyRead();
         break;
     case QSerialPort::IgnorePolicy:
@@ -663,33 +576,58 @@ void QSerialPortPrivate::emitReadyRead()
 {
     Q_Q(QSerialPort);
 
-    readyReadEmitted = true;
     emit q->readyRead();
 }
 
+qint64 QSerialPortPrivate::writeData(const char *data, qint64 maxSize)
+{
+    Q_Q(QSerialPort);
+
+    ::memcpy(writeBuffer.reserve(maxSize), data, maxSize);
+    actualBytesToWrite += maxSize;
+
+    if (!writeBuffer.isEmpty() && !writeStarted) {
+        if (!startAsyncWriteTimer) {
+            startAsyncWriteTimer = new QTimer(q);
+            q->connect(startAsyncWriteTimer, SIGNAL(timeout()), q, SLOT(_q_startAsyncWrite()));
+            startAsyncWriteTimer->setSingleShot(true);
+        }
+        startAsyncWriteTimer->start(0);
+    }
+    return maxSize;
+}
+
 void QSerialPortPrivate::handleLineStatusErrors()
 {
-    Q_Q(QSerialPort);
-
     DWORD errors = 0;
-    if (!::ClearCommError(handle, &errors, NULL)) {
-        q->setError(decodeSystemError());
+    if (!::ClearCommError(handle, &errors, Q_NULLPTR)) {
+        setError(getSystemError());
         return;
     }
 
     if (errors & CE_FRAME) {
-        q->setError(QSerialPort::FramingError);
+        setError(QSerialPortErrorInfo(QSerialPort::FramingError, QSerialPort::tr("Framing error detected while reading")));
     } else if (errors & CE_RXPARITY) {
-        q->setError(QSerialPort::ParityError);
+        setError(QSerialPortErrorInfo(QSerialPort::ParityError, QSerialPort::tr("ParityError error detected while reading")));
         parityErrorOccurred = true;
     } else if (errors & CE_BREAK) {
-        q->setError(QSerialPort::BreakConditionError);
+        setError(QSerialPortErrorInfo(QSerialPort::BreakConditionError, QSerialPort::tr("Break condition detected while reading")));
     } else {
-        q->setError(QSerialPort::UnknownError);
+        setError(QSerialPortErrorInfo(QSerialPort::UnknownError, QSerialPort::tr("Unknown streaming error")));
     }
 }
 
-inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
+OVERLAPPED *QSerialPortPrivate::waitForNotified(int msecs)
+{
+    OVERLAPPED *overlapped = notifier->waitForAnyNotified(msecs);
+    if (!overlapped) {
+        setError(getSystemError(WAIT_TIMEOUT));
+        return 0;
+    }
+    return overlapped;
+}
+
+inline bool QSerialPortPrivate::initialize()
 {
     Q_Q(QSerialPort);
 
@@ -697,7 +635,7 @@ inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
     restoredDcb.DCBlength = sizeof(restoredDcb);
 
     if (!::GetCommState(handle, &restoredDcb)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -712,11 +650,13 @@ inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
     if (currentDcb.fDtrControl ==  DTR_CONTROL_HANDSHAKE)
         currentDcb.fDtrControl = DTR_CONTROL_DISABLE;
 
+    currentDcb.BaudRate = inputBaudRate;
+
     if (!updateDcb())
         return false;
 
     if (!::GetCommTimeouts(handle, &restoredCommTimeouts)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
@@ -726,31 +666,27 @@ inline bool QSerialPortPrivate::initialize(QIODevice::OpenMode mode)
     if (!updateCommTimeouts())
         return false;
 
-    if (mode & QIODevice::ReadOnly)
-        readCompletionNotifier->setEnabled(true);
-
-    if (mode & QIODevice::WriteOnly)
-        writeCompletionNotifier->setEnabled(true);
-
     if (!::SetCommMask(handle, originalEventMask)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
 
+    notifier = new QWinOverlappedIoNotifier(q);
+    q->connect(notifier, SIGNAL(notified(quint32, quint32, OVERLAPPED*)),
+               q, SLOT(_q_notified(quint32, quint32, OVERLAPPED*)));
+    notifier->setHandle(handle);
+    notifier->setEnabled(true);
+
     if (!startAsyncCommunication())
         return false;
 
-    communicationNotifier->setEnabled(true);
-
     return true;
 }
 
 bool QSerialPortPrivate::updateDcb()
 {
-    Q_Q(QSerialPort);
-
     if (!::SetCommState(handle, &currentDcb)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
     return true;
@@ -758,124 +694,65 @@ bool QSerialPortPrivate::updateDcb()
 
 bool QSerialPortPrivate::updateCommTimeouts()
 {
-    Q_Q(QSerialPort);
-
     if (!::SetCommTimeouts(handle, &currentCommTimeouts)) {
-        q->setError(decodeSystemError());
+        setError(getSystemError());
         return false;
     }
     return true;
 }
 
-qint64 QSerialPortPrivate::handleOverlappedResult(int direction, OVERLAPPED &overlapped)
+QSerialPortErrorInfo QSerialPortPrivate::getSystemError(int systemErrorCode) const
 {
-    Q_Q(QSerialPort);
+    if (systemErrorCode == -1)
+        systemErrorCode = ::GetLastError();
 
-    DWORD bytesTransferred = 0;
-    if (!::GetOverlappedResult(handle, &overlapped, &bytesTransferred, FALSE)) {
-        const QSerialPort::SerialPortError error = decodeSystemError();
-        if (error == QSerialPort::NoError)
-            return qint64(0);
-        if (error != QSerialPort::ResourceError) {
-            if (direction == QSerialPort::Input)
-                q->setError(QSerialPort::ReadError);
-            else if (direction == QSerialPort::Output)
-                q->setError(QSerialPort::WriteError);
-            else
-                q->setError(error);
-            return qint64(-1);
-        }
-    }
-    return bytesTransferred;
-}
+    QSerialPortErrorInfo error;
+    error.errorString = qt_error_string(systemErrorCode);
 
-QSerialPort::SerialPortError QSerialPortPrivate::decodeSystemError() const
-{
-    QSerialPort::SerialPortError error;
-    switch (::GetLastError()) {
+    switch (systemErrorCode) {
+    case ERROR_SUCCESS:
+        error.errorCode = QSerialPort::NoError;
+        break;
     case ERROR_IO_PENDING:
-        error = QSerialPort::NoError;
+        error.errorCode = QSerialPort::NoError;
         break;
     case ERROR_MORE_DATA:
-        error = QSerialPort::NoError;
+        error.errorCode = QSerialPort::NoError;
         break;
     case ERROR_FILE_NOT_FOUND:
-        error = QSerialPort::DeviceNotFoundError;
+        error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
     case ERROR_INVALID_NAME:
-        error = QSerialPort::DeviceNotFoundError;
+        error.errorCode = QSerialPort::DeviceNotFoundError;
         break;
     case ERROR_ACCESS_DENIED:
-        error = QSerialPort::PermissionError;
+        error.errorCode = QSerialPort::PermissionError;
         break;
     case ERROR_INVALID_HANDLE:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
     case ERROR_INVALID_PARAMETER:
-        error = QSerialPort::UnsupportedOperationError;
+        error.errorCode = QSerialPort::UnsupportedOperationError;
         break;
     case ERROR_BAD_COMMAND:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
     case ERROR_DEVICE_REMOVED:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
         break;
     case ERROR_OPERATION_ABORTED:
-        error = QSerialPort::ResourceError;
+        error.errorCode = QSerialPort::ResourceError;
+        break;
+    case WAIT_TIMEOUT:
+        error.errorCode = QSerialPort::TimeoutError;
         break;
     default:
-        error = QSerialPort::UnknownError;
+        error.errorCode = QSerialPort::UnknownError;
         break;
     }
     return error;
 }
 
-bool QSerialPortPrivate::waitAnyEvent(int msecs, bool *timedOut, HANDLE *triggeredEvent)
-{
-    Q_Q(QSerialPort);
-
-    Q_ASSERT(timedOut);
-
-    QVector<HANDLE> handles = QVector<HANDLE>()
-            << communicationOverlapped.hEvent
-            << readCompletionOverlapped.hEvent
-            << writeCompletionOverlapped.hEvent;
-
-    DWORD waitResult = ::WaitForMultipleObjects(handles.count(),
-                                                handles.constData(),
-                                                FALSE, // wait any event
-                                                msecs == -1 ? INFINITE : msecs);
-    if (waitResult == WAIT_TIMEOUT) {
-        *timedOut = true;
-        q->setError(QSerialPort::TimeoutError, qt_error_string(WAIT_TIMEOUT));
-        return false;
-    }
-
-    if (waitResult >= DWORD(WAIT_OBJECT_0 + handles.count()))
-        return false;
-
-    *triggeredEvent = handles.at(waitResult - WAIT_OBJECT_0);
-    return true;
-}
-
-static const QString defaultPathPrefix = QStringLiteral("\\\\.\\");
-
-QString QSerialPortPrivate::portNameToSystemLocation(const QString &port)
-{
-    QString ret = port;
-    if (!ret.contains(defaultPathPrefix))
-        ret.prepend(defaultPathPrefix);
-    return ret;
-}
-
-QString QSerialPortPrivate::portNameFromSystemLocation(const QString &location)
-{
-    QString ret = location;
-    if (ret.contains(defaultPathPrefix))
-        ret.remove(defaultPathPrefix);
-    return ret;
-}
-
 // This table contains standard values of baud rates that
 // are defined in MSDN and/or in Win SDK file winbase.h
 
