@@ -1,40 +1,32 @@
 /****************************************************************************
 **
 ** Copyright (C) 2014 Denis Shienkov <denis.shienkov@gmail.com>
-** Contact: http://www.qt-project.org/legal
+** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the QtSerialPort module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:LGPL21$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see http://www.qt.io/terms-conditions. For further
+** information use the contact form at http://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 2.1 or version 3 as published by the Free
+** Software Foundation and appearing in the file LICENSE.LGPLv21 and
+** LICENSE.LGPLv3 included in the packaging of this file. Please review the
+** following information to ensure the GNU Lesser General Public License
+** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
+** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
+** As a special exception, The Qt Company gives you certain additional
+** rights. These rights are described in The Qt Company LGPL Exception
 ** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
 **
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
-**
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
@@ -43,6 +35,8 @@
 #include <QtSerialPort/QSerialPort>
 #include <QtSerialPort/QSerialPortInfo>
 
+#include <QThread>
+
 Q_DECLARE_METATYPE(QSerialPort::SerialPortError);
 Q_DECLARE_METATYPE(QIODevice::OpenMode);
 Q_DECLARE_METATYPE(QIODevice::OpenModeFlag);
@@ -64,12 +58,7 @@ public:
     static void enterLoopMsecs(int msecs)
     {
         ++loopLevel;
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
         QTestEventLoop::instance().enterLoopMSecs(msecs);
-#else
-        Q_UNUSED(msecs);
-        QTestEventLoop::instance().enterLoop(1);
-#endif
         --loopLevel;
     }
 
@@ -115,14 +104,20 @@ private slots:
     void asynchronousWriteByTimer_data();
     void asynchronousWriteByTimer();
 
+#ifdef Q_OS_WIN
+    void readBufferOverflow();
+    void readAfterInputClear();
+    void synchronousReadWriteAfterAsynchronousReadWrite();
+#endif
+
+    void controlBreak();
+
 protected slots:
     void handleBytesWrittenAndExitLoopSlot(qint64 bytesWritten);
     void handleBytesWrittenAndExitLoopSlot2(qint64 bytesWritten);
 
 private:
-#ifdef Q_OS_WIN
     void clearReceiver(const QString &customReceiverName = QString());
-#endif
 
     QString m_senderPortName;
     QString m_receiverPortName;
@@ -142,9 +137,8 @@ tst_QSerialPort::tst_QSerialPort()
 {
 }
 
-#ifdef Q_OS_WIN
 // This method is a workaround for the "com0com" virtual serial port
-// driver, which is installed on CI. The problem is that the close/clear
+// driver or for the SOCAT utility. The problem is that the close/clear
 // methods have no effect on sender serial port. If any data didn't manage
 // to be transferred before closing, then this data will continue to be
 // transferred at next opening of sender port.
@@ -160,18 +154,27 @@ void tst_QSerialPort::clearReceiver(const QString &customReceiverName)
     if (receiver.open(QIODevice::ReadOnly))
         enterLoopMsecs(100);
 }
-#endif
 
 void tst_QSerialPort::initTestCase()
 {
     m_senderPortName = QString::fromLocal8Bit(qgetenv("QTEST_SERIALPORT_SENDER"));
     m_receiverPortName = QString::fromLocal8Bit(qgetenv("QTEST_SERIALPORT_RECEIVER"));
     if (m_senderPortName.isEmpty() || m_receiverPortName.isEmpty()) {
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
-        QSKIP("Test doesn't work because the names of serial ports aren't found in env.");
+        static const char message[] =
+              "Test doesn't work because the names of serial ports aren't found in env.\n"
+              "Please set environment variables:\n"
+              " QTEST_SERIALPORT_SENDER to name of output serial port\n"
+              " QTEST_SERIALPORT_RECEIVER to name of input serial port\n"
+              "Specify short names of port"
+#if defined(Q_OS_UNIX)
+              ", like: ttyS0\n";
+#elif defined(Q_OS_WIN32)
+              ", like: COM1\n";
 #else
-        QSKIP("Test doesn't work because the names of serial ports aren't set found env.", SkipAll);
+              "\n";
 #endif
+
+        QSKIP(message);
     } else {
         m_availablePortNames << m_senderPortName << m_receiverPortName;
     }
@@ -227,9 +230,19 @@ void tst_QSerialPort::constructByName()
 void tst_QSerialPort::constructByInfo()
 {
     QSerialPortInfo senderPortInfo(m_senderPortName);
+    QSerialPortInfo receiverPortInfo(m_receiverPortName);
+
+#if defined(Q_OS_UNIX)
+    if (senderPortInfo.isNull() || receiverPortInfo.isNull()) {
+        static const char message[] =
+                "Test doesn't work because the specified serial ports aren't"
+                " found in system and can't be constructed by QSerialPortInfo.\n";
+        QSKIP(message);
+    }
+#endif
+
     QSerialPort serialPort(senderPortInfo);
     QCOMPARE(serialPort.portName(), m_senderPortName);
-    QSerialPortInfo receiverPortInfo(m_receiverPortName);
     serialPort.setPort(receiverPortInfo);
     QCOMPARE(serialPort.portName(), m_receiverPortName);
 }
@@ -311,13 +324,13 @@ void tst_QSerialPort::handleBytesWrittenAndExitLoopSlot(qint64 bytesWritten)
 void tst_QSerialPort::flush()
 {
 #ifdef Q_OS_WIN
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
     QSKIP("flush() does not work on Windows");
-#else
-    QSKIP("flush() does not work on Windows", SkipAll);
-#endif
 #endif
 
+    // the dummy device on other side also has to be open
+    QSerialPort dummySerialPort(m_receiverPortName);
+    QVERIFY(dummySerialPort.open(QIODevice::ReadOnly));
+
     QSerialPort serialPort(m_senderPortName);
     connect(&serialPort, SIGNAL(bytesWritten(qint64)), this, SLOT(handleBytesWrittenAndExitLoopSlot(qint64)));
     QSignalSpy bytesWrittenSpy(&serialPort, SIGNAL(bytesWritten(qint64)));
@@ -346,13 +359,13 @@ void tst_QSerialPort::handleBytesWrittenAndExitLoopSlot2(qint64 bytesWritten)
 void tst_QSerialPort::doubleFlush()
 {
 #ifdef Q_OS_WIN
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
     QSKIP("flush() does not work on Windows");
-#else
-    QSKIP("flush() does not work on Windows", SkipAll);
-#endif
 #endif
 
+    // the dummy device on other side also has to be open
+    QSerialPort dummySerialPort(m_receiverPortName);
+    QVERIFY(dummySerialPort.open(QIODevice::ReadOnly));
+
     QSerialPort serialPort(m_senderPortName);
     connect(&serialPort, SIGNAL(bytesWritten(qint64)), this, SLOT(handleBytesWrittenAndExitLoopSlot2(qint64)));
     QSignalSpy bytesWrittenSpy(&serialPort, SIGNAL(bytesWritten(qint64)));
@@ -388,8 +401,8 @@ void tst_QSerialPort::waitForBytesWritten()
 
 void tst_QSerialPort::waitForReadyReadWithTimeout()
 {
-#ifdef Q_OS_WIN
     clearReceiver();
+#ifdef Q_OS_WIN
     // the dummy device on other side also has to be open
     QSerialPort dummySerialPort(m_senderPortName);
     QVERIFY(dummySerialPort.open(QIODevice::WriteOnly));
@@ -470,26 +483,27 @@ void tst_QSerialPort::twoStageSynchronousLoopback()
     senderPort.waitForBytesWritten(waitMsecs);
     QTest::qSleep(waitMsecs);
     receiverPort.waitForReadyRead(waitMsecs);
-    QCOMPARE(newlineArray.size(), receiverPort.bytesAvailable());
+    QCOMPARE(receiverPort.bytesAvailable(), qint64(newlineArray.size()));
+
     receiverPort.write(receiverPort.readAll());
     receiverPort.waitForBytesWritten(waitMsecs);
     QTest::qSleep(waitMsecs);
     senderPort.waitForReadyRead(waitMsecs);
-    QCOMPARE(newlineArray.size(), senderPort.bytesAvailable());
-    QCOMPARE(newlineArray, senderPort.readAll());
+    QCOMPARE(senderPort.bytesAvailable(), qint64(newlineArray.size()));
+    QCOMPARE(senderPort.readAll(), newlineArray);
 
     // second stage
     senderPort.write(newlineArray);
     senderPort.waitForBytesWritten(waitMsecs);
     QTest::qSleep(waitMsecs);
     receiverPort.waitForReadyRead(waitMsecs);
-    QCOMPARE(newlineArray.size(), receiverPort.bytesAvailable());
+    QCOMPARE(receiverPort.bytesAvailable(), qint64(newlineArray.size()));
     receiverPort.write(receiverPort.readAll());
     receiverPort.waitForBytesWritten(waitMsecs);
     QTest::qSleep(waitMsecs);
     senderPort.waitForReadyRead(waitMsecs);
-    QCOMPARE(newlineArray.size(), senderPort.bytesAvailable());
-    QCOMPARE(newlineArray, senderPort.readAll());
+    QCOMPARE(senderPort.bytesAvailable(), qint64(newlineArray.size()));
+    QCOMPARE(senderPort.readAll(), newlineArray);
 }
 
 void tst_QSerialPort::synchronousReadWrite()
@@ -515,7 +529,7 @@ void tst_QSerialPort::synchronousReadWrite()
     while ((readData.size() < writeData.size()) && receiverPort.waitForReadyRead(100))
         readData.append(receiverPort.readAll());
 
-    QCOMPARE(writeData, readData);
+    QCOMPARE(readData, writeData);
 }
 
 class AsyncReader : public QObject
@@ -666,5 +680,235 @@ void tst_QSerialPort::asynchronousWriteByTimer()
     QCOMPARE(receiverPort.readAll(), alphabetArray);
 }
 
+#ifdef Q_OS_WIN
+void tst_QSerialPort::readBufferOverflow()
+{
+    clearReceiver();
+
+    QSerialPort senderPort(m_senderPortName);
+    QVERIFY(senderPort.open(QSerialPort::WriteOnly));
+
+    QSerialPort receiverPort(m_receiverPortName);
+    QVERIFY(receiverPort.open(QSerialPort::ReadOnly));
+
+    const int readBufferSize = alphabetArray.size() / 2;
+    receiverPort.setReadBufferSize(readBufferSize);
+    QCOMPARE(receiverPort.readBufferSize(), qint64(readBufferSize));
+
+    QCOMPARE(senderPort.write(alphabetArray), qint64(alphabetArray.size()));
+    QVERIFY2(senderPort.waitForBytesWritten(100), "Waiting for bytes written failed");
+
+    QByteArray readData;
+    while (receiverPort.waitForReadyRead(100)) {
+        QVERIFY(receiverPort.bytesAvailable() > 0);
+        readData += receiverPort.readAll();
+    }
+
+    QCOMPARE(readData, alphabetArray);
+
+    // No more bytes available
+    QVERIFY(receiverPort.bytesAvailable() == 0);
+}
+
+void tst_QSerialPort::readAfterInputClear()
+{
+    clearReceiver();
+
+    QSerialPort senderPort(m_senderPortName);
+    QVERIFY(senderPort.open(QSerialPort::WriteOnly));
+
+    QSerialPort receiverPort(m_receiverPortName);
+    QVERIFY(receiverPort.open(QSerialPort::ReadOnly));
+
+    const int readBufferSize = alphabetArray.size() / 2;
+    receiverPort.setReadBufferSize(readBufferSize);
+    QCOMPARE(receiverPort.readBufferSize(), qint64(readBufferSize));
+
+    const int waitMsecs = 100;
+
+    // First write more than read buffer size
+    QCOMPARE(senderPort.write(alphabetArray), qint64(alphabetArray.size()));
+    QVERIFY2(senderPort.waitForBytesWritten(waitMsecs), "Waiting for bytes written failed");
+
+    // Wait for first part of data into read buffer
+    while (receiverPort.waitForReadyRead(waitMsecs));
+    QCOMPARE(receiverPort.bytesAvailable(), qint64(readBufferSize));
+    // Wait for second part of data into driver's FIFO
+    QTest::qSleep(waitMsecs);
+
+    QVERIFY(receiverPort.clear(QSerialPort::Input));
+    QCOMPARE(receiverPort.bytesAvailable(), qint64(0));
+
+    // Second write less than read buffer size
+    QCOMPARE(senderPort.write(newlineArray), qint64(newlineArray.size()));
+    QVERIFY2(senderPort.waitForBytesWritten(waitMsecs), "Waiting for bytes written failed");
+
+    while (receiverPort.waitForReadyRead(waitMsecs));
+    QCOMPARE(receiverPort.bytesAvailable(), qint64(newlineArray.size()));
+    QCOMPARE(receiverPort.readAll(), newlineArray);
+
+    // No more bytes available
+    QVERIFY(receiverPort.bytesAvailable() == 0);
+}
+
+class MasterTransactor : public QObject
+{
+    Q_OBJECT
+public:
+    explicit MasterTransactor(const QString &name)
+        : serialPort(name)
+    {
+    }
+
+public slots:
+    void open()
+    {
+        if (serialPort.open(QSerialPort::ReadWrite)) {
+            createAsynchronousConnection();
+            serialPort.write("A", 1);
+        }
+    }
+
+private slots:
+    void synchronousTransaction()
+    {
+        serialPort.write("B", 1);
+        if (serialPort.waitForBytesWritten(100)) {
+            if (serialPort.waitForReadyRead(100))
+                tst_QSerialPort::exitLoop();
+        }
+    }
+
+    void transaction()
+    {
+        deleteAsyncronousConnection();
+        synchronousTransaction();
+    }
+
+private:
+    void createAsynchronousConnection()
+    {
+        connect(&serialPort, &QSerialPort::readyRead, this, &MasterTransactor::transaction);
+    }
+
+    void deleteAsyncronousConnection()
+    {
+        serialPort.disconnect();
+    }
+
+    QSerialPort serialPort;
+};
+
+class SlaveTransactor : public QObject
+{
+    Q_OBJECT
+public:
+    explicit SlaveTransactor(const QString &name)
+        : serialPort(new QSerialPort(name, this))
+    {
+        connect(serialPort, &QSerialPort::readyRead, this, &SlaveTransactor::transaction);
+    }
+
+public slots:
+    void open()
+    {
+        if (serialPort->open(QSerialPort::ReadWrite))
+            emit ready();
+    }
+
+signals:
+    void ready();
+
+private slots:
+    void transaction()
+    {
+        serialPort->write("Z", 1);
+    }
+
+private:
+    QSerialPort *serialPort;
+};
+
+void tst_QSerialPort::synchronousReadWriteAfterAsynchronousReadWrite()
+{
+    MasterTransactor master(m_senderPortName);
+    SlaveTransactor *slave = new SlaveTransactor(m_receiverPortName);
+
+    QThread thread;
+    slave->moveToThread(&thread);
+    thread.start();
+
+    QObject::connect(&thread, &QThread::finished, slave, &SlaveTransactor::deleteLater);
+    QObject::connect(slave, &SlaveTransactor::ready, &master, &MasterTransactor::open);
+
+    QMetaObject::invokeMethod(slave, "open", Qt::QueuedConnection);
+
+    tst_QSerialPort::enterLoopMsecs(500);
+
+    thread.quit();
+    thread.wait();
+
+    QVERIFY2(!timeout(), "Timed out when testing of transactions.");
+}
+
+#endif
+
+class BreakReader : public QObject
+{
+    Q_OBJECT
+public:
+    explicit BreakReader(QSerialPort &port)
+        : serialPort(port)
+    {
+        connect(&serialPort, SIGNAL(readyRead()), this, SLOT(receive()));
+    }
+
+private slots:
+    void receive()
+    {
+        tst_QSerialPort::exitLoop();
+    }
+
+private:
+    QSerialPort &serialPort;
+};
+
+void tst_QSerialPort::controlBreak()
+{
+#ifdef Q_OS_WIN
+    clearReceiver();
+#endif
+
+    QSerialPort senderPort(m_senderPortName);
+    QVERIFY(senderPort.open(QSerialPort::WriteOnly));
+    QCOMPARE(senderPort.isBreakEnabled(), false);
+
+    QSignalSpy breakSpy(&senderPort, SIGNAL(breakEnabledChanged(bool)));
+    QVERIFY(breakSpy.isValid());
+
+    QSerialPort receiverPort(m_receiverPortName);
+    QVERIFY(receiverPort.open(QSerialPort::ReadOnly));
+
+    BreakReader reader(receiverPort);
+
+    QVERIFY(senderPort.setBreakEnabled(true));
+    QCOMPARE(senderPort.isBreakEnabled(), true);
+
+    enterLoop(1);
+    QVERIFY2(!timeout(), "Timed out when waiting for the read of break state.");
+    QVERIFY(receiverPort.bytesAvailable() > 0);
+
+    foreach (const char c, receiverPort.readAll()) {
+        QCOMPARE(c, char(0));
+    }
+
+    QVERIFY(senderPort.setBreakEnabled(false));
+    QCOMPARE(senderPort.isBreakEnabled(), false);
+
+    QCOMPARE(breakSpy.count(), 2);
+    QCOMPARE(qvariant_cast<bool>(breakSpy.at(0).at(0)), true);
+    QCOMPARE(qvariant_cast<bool>(breakSpy.at(1).at(0)), false);
+}
+
 QTEST_MAIN(tst_QSerialPort)
 #include "tst_qserialport.moc"
